@@ -95,19 +95,6 @@ namespace pybind11 { namespace detail {
                 return element_to_primitive(src).release();
             }
     };
-
-    // Caster for the key_value_pairs in Object.
-    template <> struct type_caster<dom::key_value_pair> {
-        public:
-            PYBIND11_TYPE_CASTER(dom::key_value_pair, _("KeyValuePair"));
-
-            static handle cast(dom::key_value_pair src, py::return_value_policy, py::handle) {
-                return py::make_tuple(
-                    sv_to_unicode(src.key),
-                    sv_to_unicode(src.value)
-                ).release();
-            }
-    };
 }}
 
 PYBIND11_MAKE_OPAQUE(dom::array);
@@ -121,6 +108,85 @@ PYBIND11_MODULE(csimdjson, m) {
     m.attr("DEFAULT_MAX_DEPTH") = py::int_(DEFAULT_MAX_DEPTH);
     m.attr("VERSION") = py::str(STRINGIFY(SIMDJSON_VERSION));
 
+    // We can't just use py::make_iterator & py::make_key_iterator, because
+    // simdjson does not implement the expect std::iterator-like and
+    // std::pair-like properties. See upstream#1098 for future fixes.
+    // It looks tempting to replace with something like ranges-v3, which would
+    // turn all of this into a one-liner. However, ranges-v3 requires VS2019
+    // which would severely cripple our platform support.
+    struct PyKeyIterator {
+        PyKeyIterator(
+            const dom::object &obj
+        ) : obj(obj), first(obj.begin()), end(obj.end()) { }
+
+        py::object next() {
+            if (first == end)
+                throw py::stop_iteration();
+
+            auto v = first.key();
+            first++;
+            return sv_to_unicode(v);
+        }
+
+        const dom::object &obj;
+        dom::object::iterator first;
+        dom::object::iterator end;
+    };
+
+    struct PyValueIterator {
+        PyValueIterator(
+            const dom::object &obj
+        ) : obj(obj), first(obj.begin()), end(obj.end()) { }
+
+        py::object next() {
+            if (first == end)
+                throw py::stop_iteration();
+
+            auto v = first.value();
+            first++;
+            return element_to_primitive(v, true);
+        }
+
+        const dom::object &obj;
+        dom::object::iterator first;
+        dom::object::iterator end;
+    };
+
+    struct PyKeyValueIterator {
+        PyKeyValueIterator(
+            const dom::object &obj
+        ) : obj(obj), first(obj.begin()), end(obj.end()) { }
+
+        py::object next() {
+            if (first == end)
+                throw py::stop_iteration();
+
+            auto v = first.value();
+            auto k = first.key();
+            first++;
+            return py::make_tuple(
+                sv_to_unicode(k),
+                element_to_primitive(v, true)
+            );
+        }
+
+        const dom::object &obj;
+        dom::object::iterator first;
+        dom::object::iterator end;
+    };
+
+    py::class_<PyKeyIterator>(m, "KeyIterator")
+        .def("__iter__", [](PyKeyIterator &it) -> PyKeyIterator& { return it; })
+        .def("__next__", &PyKeyIterator::next);
+
+    py::class_<PyValueIterator>(m, "ValueIterator")
+        .def("__iter__", [](PyValueIterator &it) -> PyValueIterator& { return it; })
+        .def("__next__", &PyValueIterator::next);
+
+    py::class_<PyKeyValueIterator>(m, "KeyValueIterator")
+        .def("__iter__", [](PyKeyValueIterator &it) -> PyKeyValueIterator& { return it; })
+        .def("__next__", &PyKeyValueIterator::next);
+
     py::register_exception_translator([](std::exception_ptr p) {
         try {
             if (p) std::rethrow_exception(p);
@@ -133,7 +199,7 @@ PYBIND11_MODULE(csimdjson, m) {
                 case error_code::INCORRECT_TYPE:
                     // We can give better error messages in each class, this is
                     // just a catch-all.
-                    PyErr_SetString(PyExc_TypeError, "Unexpected type");
+                    PyErr_SetString(PyExc_TypeError, e.what());
                     return;
                 case error_code::MEMALLOC:
                     PyErr_SetNone(PyExc_MemoryError);
@@ -249,13 +315,15 @@ PYBIND11_MODULE(csimdjson, m) {
             "A proxy object that behaves much like a real `list()`.\n"
             "Python objects are not created until an element in the list\n"
             "is accessed.\n\n"
-            "Supports iteration, indexing, `len()`, and slicing."
+            "Complies with the `collections.abc.Sequence` interface, but does"
+            " not inherit from it."
         )
         .def("__truediv__",
             [](dom::array &self, const char *json_pointer) {
                 return element_to_primitive(self.at(json_pointer));
             },
-            py::return_value_policy::reference_internal
+            py::return_value_policy::reference_internal,
+            py::is_operator()
         )
         .def("__getitem__",
             [](dom::array &self, int64_t i) {
@@ -290,7 +358,7 @@ PYBIND11_MODULE(csimdjson, m) {
                 return result;
             }
         )
-        .def("__len__", [](dom::object &self) { return self.size(); })
+        .def("__len__", &dom::array::size)
         .def("__iter__",
             [](dom::array &self) {
                 return py::make_iterator(self.begin(), self.end());
@@ -298,12 +366,62 @@ PYBIND11_MODULE(csimdjson, m) {
             py::return_value_policy::reference_internal,
             py::keep_alive<0, 1>()
         )
+        .def("count",
+            [](dom::array &self, py::object value) {
+                // This is "correct" implementation of count at the cost
+                // of performance.
+                // We should also make a count for primitive types that
+                // converts the `value` just once, and compares against the
+                // basic types instead.
+                unsigned long long i = 0;
+                for (dom::element field : self) {
+                    if (element_to_primitive(field).equal(value)) {
+                        i++;
+                    }
+                }
+                return i;
+            }
+        )
+        .def("index",
+            [](dom::array &self, py::object x, py::object start, py::object end) {
+                // Cheat and use py::slice since it takes care of all the edge
+                // cases for us.
+                auto size = self.size();
+                size_t start_i, stop_i, step, slicelength;
+
+                py::slice slice(
+                    start.is_none() ? 0 : start.cast<ssize_t>(),
+                    end.is_none() ? size : end.cast<ssize_t>(),
+                    1
+                );
+
+                if (!slice.compute(size, &start_i, &stop_i, &step, &slicelength))
+                    throw py::error_already_set();
+
+                for (size_t i = 0; i < slicelength; ++i) {
+                    if (element_to_primitive(self.at(start_i)).equal(x)) {
+                        return start_i;
+                    }
+                    start_i += step;
+                }
+
+                throw py::value_error();
+            },
+            py::arg("x"),
+            py::arg("start") = py::none(),
+            py::arg("end") = py::none()
+        )
         .def("as_list",
             [](dom::array &self) {
                 return array_to_list(self, true);
             },
             "Convert this Array to a regular python list, recursively"
             " converting any objects/lists it finds."
+        )
+        .def_property_readonly("mini",
+            [](dom::array &self) -> std::string { return minify(self); },
+            "Returns the minified JSON representation of this Array as"
+            " a `str`."
         );
 
     py::class_<dom::object>(
@@ -312,7 +430,8 @@ PYBIND11_MODULE(csimdjson, m) {
             "A proxy object that behaves much like a real `dict()`.\n"
             "Python objects are not created until an element in the Object\n"
             "is accessed.\n\n"
-            "Supports iteration, indexing, `len()`, and `in`/`not in`."
+            "Complies with the `collections.abc.Mapping` interface, but does"
+            " not inherit from it."
         )
         .def("__truediv__",
             [](dom::object &self, const char *json_pointer) {
@@ -320,24 +439,7 @@ PYBIND11_MODULE(csimdjson, m) {
             },
             py::return_value_policy::reference_internal
         )
-        .def("at",
-            [](dom::object &self, const char *json_pointer) {
-                return element_to_primitive(self.at(json_pointer));
-            },
-            py::return_value_policy::reference_internal,
-            "Get the value associated with the given JSON pointer."
-        )
-        .def("__iter__",
-            [](dom::object &self) {
-                return py::make_iterator(
-                    self.begin(),
-                    self.end()
-                );
-            },
-            py::return_value_policy::reference_internal,
-            py::keep_alive<0, 1>()
-        )
-        .def("__len__", [](dom::object &self) { return self.size(); })
+        .def("__len__", &dom::object::size)
         .def("__getitem__",
             [](dom::object &self, const char *key) {
                 return element_to_primitive(self[key]);
@@ -350,42 +452,49 @@ PYBIND11_MODULE(csimdjson, m) {
                 return result.error() ? false : true;
             }
         )
-        .def("keys",
-            [](dom::object &self) {
-                // We can't use iterators here until upstream #1046 is fixed.
-                py::list *result = new py::list(self.size());
-                size_t i = 0;
-                for (dom::key_value_pair field : self) {
-                    PyList_SET_ITEM(
-                        result->ptr(),
-                        i,
-                        sv_to_unicode(field.key).release().ptr()
-                    );
-                    i++;
-                }
-                return result;
+        .def("__iter__",
+            [](dom::object &self) { return PyKeyIterator(self); },
+            py::return_value_policy::reference_internal,
+            py::keep_alive<0, 1>()
+        )
+        .def("at",
+            [](dom::object &self, const char *json_pointer) {
+                return element_to_primitive(self.at(json_pointer));
             },
-            "Returns a list of all keys in this `Object`."
+            py::return_value_policy::reference_internal,
+            "Get the value associated with the given JSON pointer."
+        )
+        .def("get",
+            [](dom::object &self, const char *key, py::object def) {
+                try {
+                    return element_to_primitive(self[key]);
+                } catch (const simdjson_error &e) {
+                    if (e.error() == error_code::NO_SUCH_FIELD) {
+                        return def;
+                    }
+                    throw;
+                }
+            },
+            py::arg("key"),
+            py::arg("default") = py::none(),
+            py::return_value_policy::reference_internal,
+            "Return the value of `key`, or `default` if the key does"
+            " not exist."
+        )
+        .def("keys",
+            [](dom::object &self) { return PyKeyIterator(self); },
+            "Returns an iterator over all keys in this `Object`."
         )
         .def("values",
-            [](dom::object &self) {
-                // We can't use iterators here until upstream #1046 is fixed.
-                py::list *result = new py::list(self.size());
-                size_t i = 0;
-                for (dom::key_value_pair field : self) {
-                    PyList_SET_ITEM(
-                        result->ptr(),
-                        i,
-                        element_to_primitive(
-                            field.value,
-                            true
-                        ).release().ptr()
-                    );
-                    i++;
-                }
-                return result;
-            },
-            "Returns a list of all values in this `Object`."
+            [](dom::object &self) { return PyValueIterator(self); },
+            "Returns an iterator over of all values in this `Object`."
+        )
+        .def("items",
+            [](dom::object &self) { return PyKeyValueIterator(self); },
+            py::return_value_policy::reference_internal,
+            py::keep_alive<0, 1>(),
+            "Returns an iterator over all the (key, value) pairs in this"
+            " `Object`."
         )
         .def("as_dict",
             [](dom::object &self) {
@@ -393,5 +502,10 @@ PYBIND11_MODULE(csimdjson, m) {
             },
             "Convert this `Object` to a regular python dictionary,\n"
             " recursively converting any objects or lists it finds."
+        )
+        .def_property_readonly("mini",
+            [](dom::object &self) -> std::string { return minify(self); },
+            "Returns the minified JSON representation of this Object as"
+            " a `str`."
         );
 }
